'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});
exports.BTMap = exports.defaultComparator = undefined;
exports.BPlusTree = BPlusTree;
exports.BTSet = BTSet;

var _binarysearch = require('./binarysearch');

var _binarysearch2 = _interopRequireDefault(_binarysearch);

var _nodes = require('./nodes');

var _path = require('./path');

var _path2 = _interopRequireDefault(_path);

var _utils = require('./utils');

var _constants = require('./constants');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

var binSearch = _binarysearch2.default.eq;

var NOT_FOUND = {};

var defaultComparator = exports.defaultComparator = function defaultComparator(a, b) {
    return (// eslint-disable-line no-confusing-arrow
        a === b // eslint-disable-line no-nested-ternary
        ? 0 : a < b ? -1 : 1
    );
};

// Iterator extraction funcs

var extractValue = function extractValue(leaf, idx) {
    return leaf.children[idx];
};
var extractKey = function extractKey(leaf, idx) {
    return leaf.keys[idx];
};
var extractEntry = function extractEntry(leaf, idx) {
    return [extractKey(leaf, idx), extractValue(leaf, idx)];
};

function BPlusTree(_opts) {
    var opts = _opts || {};
    this.order = _constants.ORDER;

    // Used to decide whether @@iterator returns
    // values or entries.
    this.isSet = opts.isSet === undefined ? false : opts.isSet;

    this.extractor = opts.extractor; // May be undefined on purpose.
    this.comparator = opts.comparator || defaultComparator;

    this.root = opts.root || new _nodes.Leaf({ order: this.order });
    this.size = opts.size || 0;
    this.height = opts.height || 0;
}

(0, _utils.extend)(BPlusTree.prototype, {
    has: function has(key) {
        return this.search(this.comparator, key) !== NOT_FOUND;
    },
    add: function add(value) {
        return this.set(value, value);
    },
    get: function get(key) {
        var val = this.search(this.comparator, key);
        return val === NOT_FOUND ? undefined : val;
    },
    clear: function clear() {
        return new this.constructor({
            comparator: this.comparator,
            extractor: this.extractor
        });
    },
    set: function set(key, value) {
        var cmp = this.comparator;
        var extractedKey = this.extractor ? this.extractor(key) : key;

        var result = this.root.insert(cmp, extractedKey, value);

        if (result === this.root) {
            return this;
        }

        var newRoot = void 0;
        var rootSplit = false;
        // Root was split
        if (result.length === 3) {
            rootSplit = true;
            var splitArr = result;
            var medianKey = splitArr[0];
            var splitChild = splitArr[1];
            var newChild = splitArr[2];
            newRoot = new _nodes.InternalNode({
                order: this.order,
                keys: [medianKey],
                children: [splitChild, newChild]
            });
        } else {
            newRoot = result;
        }

        return new this.constructor({
            extractor: this.extractor,
            comparator: this.comparator,
            root: newRoot,
            height: rootSplit ? this.height + 1 : this.height,
            size: this.size + 1
        });
    },
    delete: function _delete(key) {
        var newRoot = this.root.delete(this.comparator, key);
        var rootMerged = false;
        if (this.root !== newRoot) {
            if (newRoot.size < _constants.MIN_ROOT_CHILDREN) {
                var isLeaf = newRoot.constructor === _nodes.Leaf;
                if (!isLeaf) {
                    // Since the minimum number of children in
                    // the root is 2, the root must have a single
                    // child.
                    newRoot = newRoot.children[0];
                    rootMerged = true;
                }
                // If the root is a leaf, it can be empty.
            }

            return new this.constructor({
                comparator: this.comparator,
                extractor: this.extractor,
                root: newRoot,
                height: rootMerged ? this.height - 1 : this.height,
                size: this.size - 1
            });
        }
        return this;
    },
    _baseBetween: function _baseBetween(extractor, _fromKey, _toKey) {
        if (this.size === 0) return (0, _utils.getEmptyIterator)();
        var fromKey = this.extractor ? this.extractor(_fromKey) : _fromKey;

        var toKey = this.extractor ? this.extractor(_toKey) : _toKey;

        var isReverse = this.comparator(fromKey, toKey) > 0;

        var fromIsRight = isReverse;
        var toIsRight = !fromIsRight;

        var fromPath = this.findPath(fromKey, fromIsRight);
        var toPath = this.findPath(toKey, toIsRight);

        if (fromPath === null || toPath === null) {
            return (0, _utils.getEmptyIterator)();
        }

        return this._iteratorFromTo(extractor, fromPath, toPath, isReverse);
    },
    between: function between(fromKey, toKey) {
        return this._baseBetween(extractEntry, fromKey, toKey);
    },


    /**
     * Returns the internal and leaf nodes
     * on the path from root to value.
     *
     * @param  {Path} path
     * @return {Array[Node]}
     */
    _pathNodes: function _pathNodes(path) {
        var nodes = new Array(path.length);

        var curr = this.root;
        for (var i = 0; i < path.length; i++) {
            nodes[i] = curr;
            curr = curr.children[path.get(i)];
        }

        return nodes;
    },
    _nextPath: function _nextPath(path) {
        var nodes = this._pathNodes(path);
        var level = nodes.length - 1;

        // Simple case: leaf index can be incremented.
        if (path.get(level) < nodes[level].children.length - 1) {
            return path.increment(level);
        }

        level--;
        // Go through the node path starting from the lowest internal node and
        // find the first node we can increment a key for.
        while (level >= 0 && path.get(level) === nodes[level].children.length - 1) {
            level--;
        } // All nodes were at their maximum key
        if (level < 0) return null;
        return path.increment(level);
    },
    _prevPath: function _prevPath(path) {
        // Leftmost path - can't decrement.
        if (path.equals(_path2.default.EMPTY_PATH)) return null;

        var nodes = this._pathNodes(path);

        var level = path.length - 1;
        while (path.get(level) === 0) {
            level--;
        }var newPath = path.set(level, path.get(level) - 1);

        // Set all the lower keys to their
        // maximum value.
        var curr = nodes[level].children[newPath.get(level)];
        level++;
        for (; level < path.length; level++) {
            var childLen = curr.children.length;
            newPath = newPath.set(level, childLen - 1);
            curr = curr.children[childLen - 1];
        }
        return newPath;
    },
    _iterateAllWithExtractFn: function _iterateAllWithExtractFn(extractFn) {
        if (this.size === 0) return (0, _utils.getEmptyIterator)();

        return this._iteratorFromTo(extractFn, this._getLeftmostPath(), this._getRightmostPath());
    },
    _getLeafFromPath: function _getLeafFromPath(path) {
        var level = this.height;

        var currLevel = 0;
        var currNode = this.root;

        while (currLevel !== level) {
            currNode = currNode.children[path.get(currLevel++)];
        }
        return currNode;
    },
    _getRightmostPath: function _getRightmostPath() {
        if (this.size === 0) return null;

        var keypath = [];
        var curr = this.root;
        var currLevel = 0;
        var leafLevel = this.height;

        while (currLevel !== leafLevel + 1) {
            var idx = curr.children.length - 1;
            keypath.push(idx);
            curr = curr.children[idx];
            currLevel++;
        }
        return _path2.default.from(keypath);
    },
    _getLeftmostPath: function _getLeftmostPath() {
        if (this.size === 0) {
            return null;
        }

        return new _path2.default(_constants.SHIFT_LEN, this.height + 1);
    },
    _iteratorFromTo: function _iteratorFromTo(extractFn, fromPath, toPath, isReverse) {
        var _this = this;

        var getPathSuccessor = isReverse ? this._prevPath.bind(this) : this._nextPath.bind(this);

        var leafLevel = this.height;

        var currPath = fromPath;

        // Most of the time we're just getting the next
        // value from a leaf, so it makes sense to cache
        // the leaf instead of looking up the successor
        // path each time.
        var leafCacheRef = void 0;

        var done = false;
        var iterator = {
            next: function next() {
                var pathsEq = currPath !== null && currPath.equals(toPath);
                if (currPath !== null && (!pathsEq || !done)) {
                    var valIdx = currPath.get(leafLevel);
                    var leaf = leafCacheRef || _this._getLeafFromPath(currPath);
                    if (!pathsEq) {
                        if (isReverse && valIdx > 0) {
                            // Can decrement leaf index
                            currPath = currPath.decrement(leafLevel);
                        } else if (!isReverse && valIdx < leaf.children.length - 1) {
                            // Can increment leaf idx
                            currPath = currPath.increment(leafLevel);
                        } else {
                            currPath = getPathSuccessor(currPath);
                            leafCacheRef = undefined;
                        }
                    } else {
                        done = true;
                    }

                    var value = extractFn(leaf, valIdx);

                    return {
                        value: value
                    };
                }

                return {
                    done: true
                };
            }
        };

        iterator[_constants.ITERATOR_PROPNAME] = function () {
            return iterator;
        };

        return iterator;
    },
    findPath: function findPath(key, fromRight) {
        if (this.size === 0) return null;

        var cmp = this.comparator;

        var curr = this.root;
        var numAccessesToReachLeaf = this.height;

        // Path needs one more value (which key to look up in leaf)
        var keypath = new Array(numAccessesToReachLeaf + 1);

        var i = 0;
        var prev = curr;
        for (; i < numAccessesToReachLeaf; i++) {
            var _idx = curr.childIdxForKey(cmp, key);
            keypath[i] = _idx;
            prev = curr;
            curr = curr.children[_idx];
        }

        var parent = prev;

        // curr should be a leaf now.
        var idx = fromRight ? _binarysearch2.default.lte(curr.keys, key, cmp) : _binarysearch2.default.gte(curr.keys, key, cmp);

        if (idx === curr.keys.length) {
            if (!fromRight) {
                // The key we're looking for could be in the right leaf.
                keypath[i - 1]++;

                var hasRightLeaf = keypath[i - 1] < parent.children.length;
                if (!hasRightLeaf) return null;
                curr = parent.children[keypath[i - 1]];
                if (curr.keys[0] >= key) {
                    keypath[i] = 0;
                } else {
                    // out of bounds.
                    return null;
                }
            } else {
                keypath[i] = idx - 1;
            }
        } else if (idx === -1) {
            if (fromRight) {
                // the key we're looking for could be in the left leaf.
                keypath[i - 1]--;

                var hasLeftLeaf = keypath[i - 1] >= 0;
                if (!hasLeftLeaf) return null;

                curr = parent.children[keypath[i - 1]];
                if (curr.keys[curr.keys.length - 1] <= key) {
                    keypath[i] = curr.keys.length - 1;
                } else {
                    // out of bounds.
                    return null;
                }
            } else {
                keypath[i] = 0;
            }
        } else {
            keypath[i] = idx;
        }

        return _path2.default.from(keypath);
    },
    findLeaf: function findLeaf(cmp, key) {
        var curr = this.root;
        var numAccessesToReachLeaf = this.height;

        for (var i = 0; i < numAccessesToReachLeaf; i++) {
            var idx = curr.childIdxForKey(cmp, key);
            curr = curr.children[idx];
        }

        return curr;
    },
    search: function search(cmp, key) {
        var extractedKey = this.extractor ? this.extractor(key) : key;

        var leaf = this.findLeaf(cmp, extractedKey);
        var idx = binSearch(leaf.keys, extractedKey, cmp);
        if (idx === -1) {
            return NOT_FOUND;
        }

        return leaf.children[idx];
    },
    visit: function visit(fn) {
        var queue = [this.root];
        while (queue.length) {
            var curr = queue.shift();
            fn(curr);
            if (curr.constructor !== _nodes.Leaf && curr.children) {
                queue.push.apply(queue, _toConsumableArray(curr.children));
            }
        }
    }
});

var makeIteratorMethod = function makeIteratorMethod(extractor) {
    return function getRangeIterator() {
        if (arguments.length === 0) {
            return this._iterateAllWithExtractFn(extractor);
        }

        var fromKey = arguments[0];
        var toKey = arguments[1];

        return this._baseBetween(extractor, fromKey, toKey);
    };
};

BPlusTree.prototype.values = makeIteratorMethod(extractValue);
BPlusTree.prototype.valueRange = BPlusTree.prototype.values;
BPlusTree.prototype.entries = makeIteratorMethod(extractEntry);
BPlusTree.prototype.entryRange = BPlusTree.prototype.entries;
BPlusTree.prototype.keys = makeIteratorMethod(extractKey);
BPlusTree.prototype.keyRange = BPlusTree.prototype.keys;

function fromSortedBase(arr, _opts, isPairs) {
    var opts = _opts || {};

    var order = _constants.ORDER;

    // For leftmost nodes, we don't take a key
    // to their parent. That means the parent
    // could have one too few children, if it's
    // already at minimum. This adjusts for that.
    var MIN_ADJUSTMENT = 1;

    var minPageLen = _constants.LEAF_MIN_CHILDREN + MIN_ADJUSTMENT;
    var maxPageLen = _constants.LEAF_MAX_CHILDREN;

    var minNodeChildren = _constants.INTERNAL_MIN_CHILDREN;
    var maxNodeChildren = _constants.INTERNAL_MAX_CHILDREN;

    var pages = (0, _utils.boundedChunk)(minPageLen, maxPageLen, arr);

    var leafs = (0, _utils.fastMap)(function (page) {
        var leafKeys = void 0;
        if (isPairs) {
            leafKeys = (0, _utils.fastMap)(function (item) {
                return opts.extractor ? opts.extractor(item[0]) : item[0];
            }, page);
        } else {
            leafKeys = opts.extractor ? (0, _utils.fastMap)(opts.extractor, page) : page;
        }

        var leafValues = isPairs ? (0, _utils.fastMap)(function (item) {
            return item[1];
        }, page) : page;

        return new _nodes.Leaf({
            order: order,
            keys: leafKeys,
            children: leafValues
        });
    }, pages);

    var newHeight = 0;
    var newRoot = null;
    var leafsProcessed = false;
    if (leafs.length === 1) {
        newRoot = leafs[0];
    } else {
        var currLevel = leafs;

        var _loop = function _loop() {
            var chunked = (0, _utils.boundedChunk)(minNodeChildren + MIN_ADJUSTMENT, maxNodeChildren, currLevel);
            var firstInChunk = true;
            var nodes = (0, _utils.fastMap)(function (chunk) {
                // eslint-disable-line
                // For leafs, we use the first key as the key
                // for the whole chunk.
                // For internal nodes, we move the first
                // key up a level.
                var operation = leafsProcessed ? function (child) {
                    var _head = child.keys[0];
                    child.keys = (0, _utils.tail)(child.keys); // eslint-disable-line
                    return _head;
                } : function (child) {
                    return child.keys[0];
                };

                // For the left-most nodes, we don't
                // take a key from its children.
                var operateOn = firstInChunk ? (0, _utils.tail)(chunk) : chunk;

                var newKeys = (0, _utils.fastMap)(operation, operateOn);

                firstInChunk = false;
                return new _nodes.InternalNode({
                    order: order,
                    keys: newKeys,
                    children: chunk
                });
            }, chunked);

            if (nodes.length === 0) return 'break';
            currLevel = nodes;
            leafsProcessed = true;
            newHeight++;
        };

        while (currLevel.length > 1) {
            var _ret = _loop();

            if (_ret === 'break') break;
        }

        newRoot = currLevel[0];
    }

    var _constructor = this;
    return new _constructor((0, _utils.extend)({}, opts, {
        root: newRoot,
        size: arr.length,
        height: newHeight
    }));
}

function fromSortedPairs(arr, opts) {
    return fromSortedBase.call(this, arr, opts, true);
}

function fromSortedValues(arr, opts) {
    return fromSortedBase.call(this, arr, opts, false);
}

BPlusTree.from = fromSortedPairs;

BPlusTree.prototype[_constants.ITERATOR_PROPNAME] = function iterator() {
    return this.isSet ? this.values() : this.entries();
};

var BTMap = exports.BTMap = BPlusTree;

function BTSet(_opts) {
    var opts = _opts || {};
    opts.isSet = true;
    BPlusTree.call(this, opts);
}

BTSet.from = fromSortedValues;

BTSet.prototype = BPlusTree.prototype;

exports.default = BTMap;