'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});
exports.Path = exports.safePathSet = exports.clearBitRange = exports.pathSet = exports.pathGet = exports.bitSlice = exports.LEVELS = undefined;

var _constants = require('./constants');

var _utils = require('./utils');

var MIN_PATH = 0;
var NUMBER_OF_UNSIGNED_BITS = 31;

var LEVELS = exports.LEVELS = Math.floor(NUMBER_OF_UNSIGNED_BITS / _constants.SHIFT_LEN);

/**
 * Slices a range of bits from the binary
 * representation of a number.
 *
 * Example:
 *     num: decimal 376 = binary 101111000
 *     fromBit: 3
 *     toBit: 9
 *     Returns decimal 47 = binary 101111
 *
 * @param  {Number} fromBit - bit to start slicing from, inclusive
 * @param  {Number} toBit - bit to stop slicing at, exclusive
 * @param  {Number} num - number to slice
 * @return {Number} sliced number
 */
var bitSlice = exports.bitSlice = function bitSlice(fromBit, toBit, num) {
    // toBit exclusive. Mask is from bits 0..toBit - 1.
    var mask = Math.pow(2, toBit) - 1;

    return (num & mask) >>> fromBit;
};

/**
 * Gets the path value for a given level.
 *
 * Example:
 *     shiftLen: 6
 *     level: 1
 *     path:  decimal 16518336 = binary 111111 000000 110011 000000
 *     Returns: decimal 51 = 110011
 *
 * @param  {Number} shiftLen - the binary representation length of a single level
 * @param  {Number} level - the level to get a value from, starting at 0
 * @param  {Number} path - input path
 * @return {Number} value for `path` at level `level`
 */
var pathGet = exports.pathGet = function pathGet(shiftLen, level, path) {
    return bitSlice(shiftLen * level, shiftLen * (level + 1), path);
};

/**
 * Returns a new updated path where the value for a given level is
 * set to `newValue`. The current value of path at the given level must be 0.
 *
 * @param  {Number} shiftLen - the binary representation length of a single level
 * @param  {Number} level - the level to set the value to, starting at 0
 * @param  {Number} path
 * @param  {Number} newValue
 * @return {Number} an updated path
 */
var pathSet = exports.pathSet = function pathSet(shiftLen, level, path, newValue) {
    return path | newValue << shiftLen * level;
};

var clearBitRange = exports.clearBitRange = function clearBitRange(fromBit, toBit, num) {
    var rangeLen = toBit - fromBit;
    var mask = Math.pow(2, rangeLen) - 1;
    return num & ~(mask << fromBit);
};

// Safe version of `pathSet` where the current value is cleared before
// assigning a new value.
var safePathSet = exports.safePathSet = function safePathSet(shiftLen, level, path, newValue) {
    var cleared = clearBitRange(shiftLen * level, shiftLen * (level + 1), path);
    return pathSet(shiftLen, level, cleared, newValue);
};

function Path(shiftLen, levels, path) {
    this.shiftLen = shiftLen || _constants.SHIFT_LEN;
    this.length = levels || LEVELS;
    this._path = path || 0;
}

Path.EMPTY_PATH = new Path(_constants.SHIFT_LEN, LEVELS, MIN_PATH);
Path.from = function from(arr) {
    var levels = arr.length;

    var path = 0;
    for (var i = 0; i < levels; i++) {
        path = pathSet(_constants.SHIFT_LEN, i, path, arr[i]);
    }

    return new Path(_constants.SHIFT_LEN, levels, path);
};

(0, _utils.extend)(Path.prototype, {
    get: function get(level) {
        return pathGet(this.shiftLen, level, this._path);
    },
    equals: function equals(otherPath) {
        return this._path === otherPath._path;
    },
    clearAfter: function clearAfter(level) {
        var currLevel = level + 1;
        var newPath = this;
        while (currLevel < this.length) {
            newPath = newPath.set(currLevel, 0);
            currLevel++;
        }
        return newPath;
    },
    toArray: function toArray() {
        var arr = new Array(this.length);
        for (var i = 0; i < this.length; i++) {
            arr[i] = this.get(i);
        }
        return arr;
    },
    compareTo: function compareTo(otherPath) {
        var a = this._path;
        var b = otherPath._path;
        return a === b ? 0 : a < b ? -1 : 1; // eslint-disable-line no-nested-ternary
    },
    increment: function increment(level) {
        var newPath = this.set(level, this.get(level) + 1);
        return newPath.clearAfter(level);
    },
    decrement: function decrement(level) {
        return this.set(level, this.get(level) - 1);
    },
    set: function set(level, value) {
        var newPath = safePathSet(this.shiftLen, level, this._path, value);
        return new Path(this.shiftLen, this.length, newPath);
    }
});

exports.Path = Path;
exports.default = Path;