'use strict';

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _chai = require('chai');

var _chai2 = _interopRequireDefault(_chai);

var _sinon = require('sinon');

var _sinon2 = _interopRequireDefault(_sinon);

var _sinonChai = require('sinon-chai');

var _sinonChai2 = _interopRequireDefault(_sinonChai);

var _index = require('../index');

var _nodes = require('../nodes');

var _path = require('../path');

var _constants = require('../constants');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

_chai2.default.use(_sinonChai2.default);
var expect = _chai2.default.expect;


function assertEmptyIterator(iterator) {
    expect(iterator.next).to.be.a.function;
    expect(iterator.next()).to.deep.equal({ done: true });
}

describe('BTree', function () {
    describe('constructor', function () {
        it('works without arguments', function () {
            var tree = new _index.BTMap();
            expect(tree.order).to.equal(_constants.ORDER);
            expect(tree.comparator).to.equal(_index.defaultComparator);
            expect(tree.size).to.equal(0);
            expect(tree.height).to.equal(0);
            expect(tree.extractor).to.be.undefined;

            expect(tree.root).to.be.an.instanceOf(_nodes.Leaf);
            expect(tree.root.order).to.equal(_constants.ORDER);
        });

        it('works with a custom comparator', function () {
            var cmp = function cmp(a, b) {
                return 0;
            };
            var tree = new _index.BTMap({ comparator: cmp });
            expect(tree.comparator).to.equal(cmp);
        });

        it('works with a custom extractor', function () {
            var extractor = function extractor(item) {
                return item.value;
            };
            var tree = new _index.BTMap({ extractor: extractor });
            expect(tree.extractor).to.equal(extractor);
        });
    });

    describe('methods', function () {
        var tree = void 0;

        beforeEach(function () {
            tree = _index.BTMap.from([[1, 1], [2, 2], [3, 3]]);
        });

        it('has', function () {
            var searchSpy = _sinon2.default.spy(tree, 'search');

            expect(tree.has(0)).to.be.false;
            expect(searchSpy).to.have.been.calledWith(tree.comparator, 0);

            expect(tree.has(1)).to.be.true;
            expect(tree.has(2)).to.be.true;
            expect(tree.has(3)).to.be.true;
            expect(tree.has(4)).to.be.false;
        });

        it('get', function () {
            var searchSpy = _sinon2.default.spy(tree, 'search');

            expect(tree.get(0)).to.be.undefined;

            expect(searchSpy).to.have.been.calledWith(tree.comparator, 0);

            expect(tree.get(1)).to.equal(1);
            expect(tree.get(2)).to.equal(2);
            expect(tree.get(3)).to.equal(3);
            expect(tree.get(4)).to.be.undefined;
        });

        it('add', function () {
            var setSpy = _sinon2.default.spy(tree, 'set');

            var initialTree = tree;

            tree = tree.add(4);
            expect(setSpy).to.have.been.calledWith(4, 4);
            expect(tree).to.not.equal(initialTree);
            expect(tree.size).to.equal(4);
        });

        it('set', function () {
            var withFour = tree.set(4, 4);
            expect(withFour).not.to.equal(tree);
            expect(withFour.size).to.equal(tree.size + 1);
            expect(withFour.height).to.equal(tree.height);
            expect(withFour.get(4)).to.equal(4);
        });

        it('set with root splitting', function () {
            var insertStub = _sinon2.default.stub(tree.root, 'insert');
            var risingKey = 'key';
            var leftChild = {};
            var rightChild = {};
            insertStub.returns([risingKey, leftChild, rightChild]);

            var withFour = tree.set(4, 4);
            expect(insertStub).to.have.been.calledWith(tree.comparator, 4, 4);
            expect(withFour.height).to.equal(tree.height + 1);
            expect(withFour.size).to.equal(tree.size + 1);

            expect(withFour.root.keys[0]).to.equal(risingKey);
            expect(withFour.root.children[0]).to.equal(leftChild);
            expect(withFour.root.children[1]).to.equal(rightChild);
        });

        it('set without modifications', function () {
            var insertStub = _sinon2.default.stub(tree.root, 'insert');
            insertStub.returns(tree.root);

            var sameTree = tree.set(3, 3);
            expect(sameTree).to.equal(tree);
        });

        it('clear', function () {
            var cleared = tree.clear();
            expect(cleared.size).to.equal(0);
            expect(cleared.height).to.equal(0);
            expect(cleared.root).not.to.equal(tree.root);
        });

        it('clear empty tree', function () {
            var emptyTree = new _index.BTMap();
            var cleared = emptyTree.clear();
            expect(cleared.size).to.equal(0);
            expect(cleared.height).to.equal(0);
            expect(cleared.root).not.to.equal(tree.root);
        });

        it('delete', function () {
            var deleteSpy = _sinon2.default.spy(tree.root, 'delete');

            var withoutTwo = tree.delete(2);

            expect(deleteSpy).to.have.been.calledWith(tree.comparator, 2);
            expect(withoutTwo).to.not.equal(tree);
            expect(withoutTwo.size).to.equal(tree.size - 1);
            expect(withoutTwo.height).to.equal(tree.height);
        });

        it('delete to empty tree', function () {
            tree = _index.BTMap.from([[1, 1]]);
            var deleteSpy = _sinon2.default.spy(tree.root, 'delete');

            var empty = tree.delete(1);
            expect(deleteSpy).to.have.been.calledWith(tree.comparator, 1);

            expect(empty.size).to.equal(0);
            expect(empty.height).to.equal(tree.height);
        });

        it('delete with collapsing root', function () {
            var rootNode = new _nodes.InternalNode({
                keys: [2],
                children: [new _nodes.Leaf({
                    keys: [1],
                    children: [1]
                }), new _nodes.Leaf({
                    keys: [2],
                    children: [2]
                })]
            });

            tree = new _index.BTMap({ root: rootNode, height: 1, size: 2 });

            var collapsed = tree.delete(1);

            expect(collapsed.root).to.be.an.instanceOf(_nodes.Leaf);
            expect(collapsed.root.keys).to.deep.equal([2]);
            expect(collapsed.root.children).to.deep.equal([2]);
        });

        it('delete without modification', function () {
            var sameTree = tree.delete(999);
            expect(sameTree).to.equal(tree);
        });

        it('iterable protocol', function () {
            expect(tree[Symbol.iterator]).to.be.a('function');
            var iter = tree[Symbol.iterator]();
            expect(iter[Symbol.iterator]).to.be.a('function');
            expect(iter.next).to.be.a('function');

            expect(Array.from(iter)).to.deep.equal([[1, 1], [2, 2], [3, 3]]);
            expect(Array.from(tree)).to.deep.equal([[1, 1], [2, 2], [3, 3]]);
        });

        it('keys', function () {
            var iterator = tree.keys();
            expect(iterator.next).to.be.a('function');
            expect(iterator[Symbol.iterator]).to.be.a('function');

            expect(Array.from(iterator)).to.deep.equal([1, 2, 3]);
        });

        it('keys on empty tree', function () {
            var emptyTree = tree.clear();
            var iterator = emptyTree.keys();
            expect(Array.from(iterator)).to.deep.equal([]);
        });

        it('values', function () {
            var iterator = tree.values();
            expect(iterator.next).to.be.a('function');
            expect(iterator[Symbol.iterator]).to.be.a('function');

            expect(Array.from(iterator)).to.deep.equal([1, 2, 3]);
        });

        it('entries', function () {
            var iterator = tree.entries();
            expect(iterator.next).to.be.a('function');
            expect(iterator[Symbol.iterator]).to.be.a('function');
            expect(Array.from(iterator)).to.deep.equal([[1, 1], [2, 2], [3, 3]]);
        });

        describe('methods related to paths', function () {
            beforeEach(function () {
                var rootNode = new _nodes.InternalNode({
                    keys: [2],
                    children: [new _nodes.Leaf({
                        keys: [1],
                        children: [1]
                    }), new _nodes.Leaf({
                        keys: [2, 3],
                        children: [2, 3]
                    })]
                });

                tree = new _index.BTMap({ root: rootNode, height: 1, size: 2 });
            });

            it('_pathNodes', function () {
                var nodes = tree._pathNodes(_path.Path.from([0, 0]));
                expect(nodes).to.have.lengthOf(2);
                expect(nodes[0]).to.equal(tree.root);
                expect(nodes[1]).to.equal(tree.root.children[0]);

                var nodes2 = tree._pathNodes(_path.Path.from([0, 1]));
                expect(nodes2).to.have.lengthOf(2);
                expect(nodes2[0]).to.equal(tree.root);
                expect(nodes2[1]).to.equal(tree.root.children[0]);

                var nodes3 = tree._pathNodes(_path.Path.from([1, 0]));
                expect(nodes3).to.have.lengthOf(2);
                expect(nodes3[0]).to.equal(tree.root);
                expect(nodes3[1]).to.equal(tree.root.children[1]);
            });

            it('_getLeftmostPath, null case', function () {
                var empty = tree.clear();
                expect(empty._getLeftmostPath()).to.be.null;
            });

            it('_getLeftmostPath', function () {
                var path = tree._getLeftmostPath();
                expect(path.toArray()).to.deep.equal([0, 0]);
            });

            it('_getRightmostPath', function () {
                var path = tree._getRightmostPath();
                expect(path.toArray()).to.deep.equal([1, 1]);
            });

            it('_getRightmostPath, null case', function () {
                var path = tree.clear()._getRightmostPath();
                expect(path).to.be.null;
            });

            it('_getLeafFromPath', function () {
                var path = _path.Path.from([0, 0]);
                var leaf = tree._getLeafFromPath(path);
                expect(leaf).to.equal(tree.root.children[0]);

                var path2 = _path.Path.from([1, 0]);
                var leaf2 = tree._getLeafFromPath(path2);
                expect(leaf2).to.equal(tree.root.children[1]);
            });

            it('between', function () {
                var findPathSpy = _sinon2.default.spy(tree, 'findPath');
                var getIteratorSpy = _sinon2.default.spy(tree, '_iteratorFromTo');

                var result = tree.between(2, 3);
                expect(result).to.equal(getIteratorSpy.firstCall.returnValue);

                var findFromPathCall = findPathSpy.getCall(0);
                var findToPathCall = findPathSpy.getCall(1);

                var fromLeft = false;
                var fromRight = true;

                expect(findPathSpy).to.have.been.calledWith(2, fromLeft);
                expect(findPathSpy).to.have.been.calledWith(3, fromRight);

                var iteratorCall = getIteratorSpy.getCall(0);

                var _iteratorCall$args = _slicedToArray(iteratorCall.args, 4);

                var _ = _iteratorCall$args[0];
                var fromPath = _iteratorCall$args[1];
                var toPath = _iteratorCall$args[2];
                var isReverse = _iteratorCall$args[3];

                expect(fromPath).to.equal(findFromPathCall.returnValue);
                expect(toPath).to.equal(findToPathCall.returnValue);
                expect(isReverse).to.be.false;
            });

            it('between same value', function () {
                var findPathSpy = _sinon2.default.spy(tree, 'findPath');
                var getIteratorSpy = _sinon2.default.spy(tree, '_iteratorFromTo');

                var result = tree.between(2, 2);
                expect(result).to.equal(getIteratorSpy.firstCall.returnValue);

                var findFromPathCall = findPathSpy.getCall(0);
                var findToPathCall = findPathSpy.getCall(1);

                var fromLeft = false;
                var fromRight = true;

                expect(findPathSpy).to.have.been.calledWith(2, fromLeft);
                expect(findPathSpy).to.have.been.calledWith(2, fromRight);

                var iteratorCall = getIteratorSpy.getCall(0);

                var _iteratorCall$args2 = _slicedToArray(iteratorCall.args, 4);

                var _ = _iteratorCall$args2[0];
                var fromPath = _iteratorCall$args2[1];
                var toPath = _iteratorCall$args2[2];
                var isReverse = _iteratorCall$args2[3];

                expect(fromPath).to.equal(findFromPathCall.returnValue);
                expect(toPath).to.equal(findToPathCall.returnValue);
                expect(isReverse).to.be.false;
            });

            it('between nonexisting value', function () {
                var result = tree.between(999, 1000);
                assertEmptyIterator(result);
            });

            it('between reverse', function () {
                var findPathSpy = _sinon2.default.spy(tree, 'findPath');
                var getIteratorSpy = _sinon2.default.spy(tree, '_iteratorFromTo');

                var result = tree.between(3, 2);
                expect(result).to.equal(getIteratorSpy.firstCall.returnValue);

                var findFromPathCall = findPathSpy.firstCall;
                var findToPathCall = findPathSpy.secondCall;

                var fromLeft = false;
                var fromRight = true;

                expect(findPathSpy).to.have.been.calledWith(3, fromRight);
                expect(findPathSpy).to.have.been.calledWith(2, fromLeft);

                var iteratorCall = getIteratorSpy.getCall(0);

                var _iteratorCall$args3 = _slicedToArray(iteratorCall.args, 4);

                var _ = _iteratorCall$args3[0];
                var fromPath = _iteratorCall$args3[1];
                var toPath = _iteratorCall$args3[2];
                var isReverse = _iteratorCall$args3[3];


                expect(fromPath).to.equal(findFromPathCall.returnValue);
                expect(toPath).to.equal(findToPathCall.returnValue);
                expect(isReverse).to.be.true;
            });

            it('findPath', function () {
                var isLeft = undefined;
                var isRight = true;

                var path = tree.findPath(1, isLeft);
                expect(path.toArray()).to.deep.equal([0, 0]);

                var nonexistingPath = tree.findPath(0, isLeft);
                // Should return the leftmost path in this case.
                expect(nonexistingPath.toArray()).to.deep.equal([0, 0]);

                expect(tree.findPath(0, isRight)).to.be.null;

                var betweenKeys = tree.findPath(1.5, isLeft);
                expect(betweenKeys.toArray()).to.deep.equal([1, 0]);

                var betweenKeys2 = tree.findPath(1.5, isRight);
                expect(betweenKeys2.toArray()).to.deep.equal([0, 0]);

                var nonexistingPath2 = tree.findPath(4, isRight);
                expect(nonexistingPath2.toArray()).to.deep.equal([1, 1]);

                expect(tree.findPath(4, isLeft)).to.be.null;
            });

            it('_nextPath', function () {
                var leftPath = tree._getLeftmostPath();

                var pathSequence = [];
                var currPath = leftPath;
                while (currPath !== null) {
                    pathSequence.push(currPath);
                    currPath = tree._nextPath(currPath);
                }
                expect(pathSequence.map(function (path) {
                    return path.toArray();
                })).to.deep.equal([[0, 0], [1, 0], [1, 1]]);
            });

            it('_prevPath', function () {
                var rightPath = tree._getRightmostPath();

                var pathSequence = [];
                var currPath = rightPath;
                while (currPath !== null) {
                    pathSequence.push(currPath);
                    currPath = tree._prevPath(currPath);
                }
                expect(pathSequence.map(function (path) {
                    return path.toArray();
                })).to.deep.equal([[1, 1], [1, 0], [0, 0]]);
            });
        });
    });
});

describe('BTSet', function () {
    it('constructor', function () {
        var set = new _index.BTSet();
        expect(set.isSet).to.be.true;
    });

    it('iterator', function () {
        var set = _index.BTSet.from([1, 2, 3]);
        var result = Array.from(set);
        expect(result).to.deep.equal([1, 2, 3]);
    });
});