'use strict';

var _head = require('ramda/src/head');

var _head2 = _interopRequireDefault(_head);

var _range = require('ramda/src/range');

var _range2 = _interopRequireDefault(_range);

var _chai = require('chai');

var _chai2 = _interopRequireDefault(_chai);

var _sinonChai = require('sinon-chai');

var _sinonChai2 = _interopRequireDefault(_sinonChai);

var _random = require('lodash/random');

var _random2 = _interopRequireDefault(_random);

var _constants = require('../constants');

var _index = require('../index');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

_chai2.default.use(_sinonChai2.default);
var expect = _chai2.default.expect;


var ASSERT_MARGIN = 1;

function assertTreeOrder(order, tree) {
    if (tree.size < order) return;

    var rootVisited = false;
    tree.visit(function (node) {
        if (!rootVisited) {
            if (node.children) {
                expect(node.children.length).to.be.at.least(2);
                expect(node.children.length).to.be.at.most(_constants.INTERNAL_MAX_CHILDREN);
            }
            rootVisited = true;
            return;
        }
        if (node.children) {
            expect(node.children.length).to.be.at.least(_constants.INTERNAL_MIN_CHILDREN - ASSERT_MARGIN, 'internal min children check failed');
            expect(node.children.length).to.be.at.most(_constants.INTERNAL_MAX_CHILDREN, 'internal max children check failed');
        } else {
            expect(node.values.length).to.be.at.least(_constants.LEAF_MIN_CHILDREN - ASSERT_MARGIN, 'leaf min children check failed');
            expect(node.values.length).to.be.at.most(_constants.LEAF_MAX_CHILDREN, 'leaf max children check failed');
        }
    });
}

var testSizes = (0, _range2.default)(0, 500);
var loops = (0, _range2.default)(0, 10);

loops.forEach(function () {
    if (!testSizes.length) return;
    var testSizeRanIdx = (0, _random2.default)(0, testSizes.length - 1);

    var testSize = testSizes[testSizeRanIdx];
    testSizes.splice(testSizeRanIdx, 1);

    describe('Integration test. Size: ' + testSize, function () {
        describe('bplus Tree', function () {
            it('init', function () {
                var tree = new _index.BTMap();
                var _tree = tree;
                for (var i = 0; i < testSize; i++) {
                    assertTreeOrder(tree.order, tree);
                    _tree = _tree.set(i, 'value' + i);
                }

                assertTreeOrder(tree.order, tree);

                for (var _i = 0; _i < testSize; _i++) {
                    expect(_tree.has(_i)).to.be.true;
                    expect(_tree.get(_i)).to.equal('value' + _i, 'get ' + _i);
                }

                expect(_tree.size).to.equal(testSize);
            });

            it('specifying extractor works', function () {
                var data = new Array(testSize);
                for (var i = 0; i < testSize; i++) {
                    var item = { value: i };
                    data[i] = [item, item];
                }

                var tree = _index.BTMap.from(data, {
                    extractor: function extractor(item) {
                        return item.value;
                    }
                });

                for (var _i2 = 0; _i2 < testSize; _i2++) {
                    expect(tree.has({ value: _i2 })).to.be.true;
                }
            });

            it('bulkload works', function () {
                var data = new Array(testSize);
                for (var i = 0; i < testSize; i++) {
                    data[i] = [i, 'value' + i];
                }

                var tree = _index.BTMap.from(data);

                assertTreeOrder(tree.order, tree);
                for (var _i3 = 0; _i3 < testSize; _i3++) {
                    expect(tree.get(_i3)).to.equal('value' + _i3);
                }
                expect(tree.size).to.equal(testSize);
            });

            describe('iteration', function () {
                var data = [];
                for (var i = 0; i < testSize; i++) {
                    data.push([i, 'value' + i]);
                }
                var tree = void 0;

                beforeEach(function () {
                    tree = _index.BTMap.from(data);
                });

                it('correctly iterates', function () {
                    var iter = tree.values();
                    var curr = iter.next();

                    if (testSize === 0) {
                        expect(curr.done).to.be.true;
                    }

                    var i = 0;
                    while (i < testSize) {
                        expect(curr.value).to.equal('value' + i);
                        i++;
                        curr = iter.next();
                    }
                    expect(curr.done).to.be.true;
                    var n = 0;
                    var _iteratorNormalCompletion = true;
                    var _didIteratorError = false;
                    var _iteratorError = undefined;

                    try {
                        for (var _iterator = tree[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                            var val = _step.value;

                            expect(val).to.deep.equal([n, 'value' + n]);
                            n++;
                        }
                    } catch (err) {
                        _didIteratorError = true;
                        _iteratorError = err;
                    } finally {
                        try {
                            if (!_iteratorNormalCompletion && _iterator.return) {
                                _iterator.return();
                            }
                        } finally {
                            if (_didIteratorError) {
                                throw _iteratorError;
                            }
                        }
                    }
                });
            });

            describe('range search', function () {
                it('works', function () {
                    var data = [];
                    for (var i = 0; i < testSize; i++) {
                        data.push([i, 'value' + i]);
                    }

                    var tree = _index.BTMap.from(data);

                    assertTreeOrder(tree.order, tree);

                    if (testSize === 0) {
                        expect(Array.from(tree.between(0, 10)).length).to.equal(0);
                    } else {
                        var end = testSize - 2;
                        var start = 1;
                        if (end < start) {
                            end = start;
                        }
                        var n = end - start + 1;
                        var result = Array.from(tree.between(start, end));
                        expect(result.length).to.equal(n);
                    }
                });

                it('works when boundaries are outside keys', function () {
                    var data = [];
                    for (var i = 0; i < testSize; i++) {
                        data.push([i, 'value' + i]);
                    }

                    var tree = _index.BTMap.from(data);

                    assertTreeOrder(tree.order, tree);

                    var result = Array.from(tree.between(-10, testSize + 10));
                    expect(result).to.have.length(testSize);
                });

                it('works in reverse when boundaries are outside keys', function () {
                    var data = [];
                    for (var i = 0; i < testSize; i++) {
                        data.push([i, 'value' + i]);
                    }

                    var tree = _index.BTMap.from(data);

                    assertTreeOrder(tree.order, tree);

                    var result = Array.from(tree.between(testSize + 10, -10));
                    expect(result).to.have.length(testSize);
                });

                it('both values outside key range returns empty iterator', function () {
                    var data = [];
                    for (var i = 0; i < testSize; i++) {
                        data.push([i, 'value' + i]);
                    }

                    var tree = _index.BTMap.from(data);

                    assertTreeOrder(tree.order, tree);

                    var result = Array.from(tree.between(testSize + 10, testSize + 15));
                    expect(result).to.have.lengthOf(0);
                });

                it('works in reverse', function () {
                    var data = [];
                    for (var i = 0; i < testSize; i++) {
                        data.push([i, 'value' + i]);
                    }

                    var tree = _index.BTMap.from(data);

                    assertTreeOrder(tree.order, tree);

                    if (testSize === 0) {
                        expect(Array.from(tree.between(10, 0)).length).to.equal(0);
                    } else {
                        var end = testSize - 2;
                        var start = 1;
                        if (end < start) {
                            end = start;
                        }
                        var n = end - start + 1;
                        var result = Array.from(tree.between(start, end));
                        var reverseResult = Array.from(tree.between(end, start));

                        var _resultReversed = result.slice();
                        _resultReversed.reverse();
                        expect(_resultReversed).to.deep.equal(reverseResult);
                        expect(result.length).to.equal(n);
                        expect(reverseResult.length).to.equal(n);
                    }
                });
            });

            describe('delete', function () {
                this.timeout(5000);

                var data = new Array(testSize);
                for (var i = 0; i < testSize; i++) {
                    data[i] = [i, 'value' + i];
                }
                var tree = void 0;

                beforeEach(function () {
                    tree = _index.BTMap.from(data);
                });

                it('correctly deletes', function () {
                    var keysToDelete = data.map(_head2.default);
                    keysToDelete.forEach(function (key) {
                        assertTreeOrder(tree.order, tree);
                        tree = tree.delete(key);
                    });
                    expect(tree.root.size).to.equal(0);
                    expect(tree.size).to.equal(0);
                });

                it('correctly deletes from right to left', function () {
                    var keysToDelete = data.map(_head2.default).reverse();

                    keysToDelete.forEach(function (key) {
                        assertTreeOrder(tree.order, tree);
                        tree = tree.delete(key);
                    });
                    expect(tree.root.size).to.equal(0);
                    expect(tree.size).to.equal(0);
                });

                it('correctly deletes on random delete order', function () {
                    var keysToDelete = data.map(_head2.default);
                    while (keysToDelete.length) {
                        assertTreeOrder(tree.order, tree);
                        var randomIdx = (0, _random2.default)(0, keysToDelete.length - 1);
                        var key = keysToDelete[randomIdx];
                        keysToDelete.splice(randomIdx, 1);
                        tree = tree.delete(key);
                    }
                    expect(tree.root.size).to.equal(0);
                    expect(tree.size).to.equal(0);
                });
            });
        });
    });
});